local ALPHA = LibStub("AceAddon-3.0"):NewAddon("ALPHA", "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")
-- Also going to need AceTimer. Leaving it out for now, though.
--local L = LibStub("AceLocale-3.0"):GetLocale("ALPHA")

--local LDB = LibStub:GetLibrary("LibDataBroker-1.1")
local Prism = LibStub("LibPrism-1.0")

-- Sometimes I hate the API.
local RANDOM_ROLL_RESULT = "^" .. string.gsub(string.gsub(string.gsub(string.gsub(RANDOM_ROLL_RESULT, "%(", "%%("), "%)", "%%)"), "%%s", "(.+)"), "%%d", "(%%d+)") .. "$"

-- The below ones may be useful. Keep for now.
local ERR_GUILD_LEAVE_S = "^" .. string.gsub(ERR_GUILD_LEAVE_S, "%%s", "(.+)") .. "$"
local ERR_LEFT_GROUP_S = "^" .. string.gsub(ERR_LEFT_GROUP_S, "%%s", "(.+)") .. "$"
--ERR_LEFT_GROUP_YOU, does not need any substitution.

local SEP = "\7" -- Beep. Could use another character as a separator as well, but eh, beeps are fun.

local blargh = {"Murlocs!", "Cookie!", "Ninja", "Forceloot"}
local myLP = 0
local current_roll = nil
local rolls = {}
local candidate_list = {} -- Index by name, include information about luck points and which slots the person is loot eligible for.
local loot = {
   slots = {}
}
local update_found = nil

local lootOpen = false

local defaults = {
   private = {
      anchor = {
	 x = 500,
	 y = -200,
      },
      rolltime = 15,
   },
   guild = {
      award = 5,
      cost = 10,
      maxlp = 200,
   }
}

if not ALPHACFG or type(ALPHACFG) ~= "table" then ALPHACFG = {} end
local cfg = {}

local mt = {
   __index = function (_, key)
      return defaults[key]
   end
   __newindex = function(t,k,v)
      rawset(t,k,v)
      if type(v) == "table" then
	 setmetatable(v, {__index = function(_,key) return defaults[k][key] end} -- Kind of funky code. A bit hacky feel. But it works for now. Look over and improve later in case it's needed.
      end
   end
}

setmetatable(cfg, mt)
cfg.private = ALPHACFG
cfg.guild = {}

local gui = nil

local useLP = true

-- Local functions!
local function getVersion()
   local version = GetAddOnMetadata("ALPHA", "Version")
   local major,minor = string.match(version, "(%d+).(%d+)") -- In some cases, we don't want crap tossed at us by alpha/beta versions. Just numbahs.
   return version,major and major or 0,minor and minor or 0
end

local function getLuckPoints(character)
   if not UnitIsInMyGuild(character) or not CanViewOfficerNote() then return 0 end
   local fname,frealm = UnitFullName(character)
   for i = 1, GetNumGuildMembers() do
      local name,_,_,_,_,_,_,officerNote = GetGuildRosterInfo(i)
      --if character == name or character == Ambiguate(name, "guild") then
      if fname.."-"..frealm == name or fname == name then -- This should be enough to properly catch it.
	 local _,points,timestamp = string.match(officerNote, "ALPHA:([%d%.]+):(%d+):(%d+)")
	 if not points then points = 0 end
	 return points,i,timestamp
      end
   end
   ALPHA:Print("WARNING: Failed to fetch luck points for loot candidate", character)
   return nil -- For some reason we failed.
end

function modLuckPoints(character, addend)
   if not UnitIsInMyGuild(character) or not CanEditOfficerNote() or not CanViewOfficerNote() then return false end
   local points,pos,_ = getLuckPoints(character)
   if not points then return false end
   local _,major,minor = getVersion()
   GuildRosterSetOfficerNote(i, string.format("ALPHA:%d.%d:%d:%d", major, minor, points + addend < 0 and 0 or points + addend > cfg.guild.maxlp and cfg.guild.maxlp or points + addend, time())) -- So sexy.
   return true
end

local function parseLootLink(lootlink)
   local color,itemID,_,_,_,_,_,suffixID,_,_,_,instanceDiffID,bonus,name = string.match(lootlink, "^|c(%x+)|Hitem:(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(-*%d+):(%d+):(%d+):(%d+):(%d+):([%d:]+)|h%[(.+)%]|h|r$")

   if not color or not itemID or not suffixID or not instanceDiffID or not bonus or not name then
      ALPHA:Print("parseLootLink(): failed to read the itemlink for " .. lootlink .. ", please harass Aiue/Gyrm about this, then go eat a cookie.")
      ALPHA:Print(string.gsub(lootlink, "|", "\\"))
      return nil
   end

   local warforged = false
   local socketed = false
   local tertiary = nil
   local first = false

   if bonus ~= "0" then
      for _,v in ipairs({string.split(":", bonus)}) do
	 if not first then -- just the number of bonuses
	    first = true
	    if v == 561 then warforged = true
	    elseif v == 565 then socketed = true
	    elseif v == 40 or v == 41 or v == 42 or v == 43 then tertiary = v
	    elseif v ~= 566 or v ~= 567 then -- Just heroic/mythic. Don't need to keep track of these.
	       ALPHA:Print("Unknown bonus id",v,bonus)
	       ALPHA:Print(string.gsub(lootlink, "|", "\\"))
	    end
	 end
      end
   end

   local link = string.format("%d:%d:%d:%s", itemID, suffixID, instanceDiffID, bonus)
   return link,name,itemID,color,warforged,socketed,tertiary
end

local function guildCfgGet(settings)
   local cfg = cfg.guild
   local set
   if not settings then
      local found = false
      local guildInfoText = GetGuildInfoText()
      if not guildInfoText then return end
      lines = {strsplit("\n", guildInfoText)}
      for k,v in pairs(lines) do -- Order does not matter.
	 if string.match(v, "^ALPHA:(%d+).(%d):(.+):END") then
	    found = true
	    set = strsplit(":", v)
	    break
	 end
      end
      if not found then return end
   else
      set = {strsplit("\7", settings)}
   end
   -- This means set[1] will be the "ALPHA", telling us we're starting, set[2] will be the version.
   -- The format for this string is expected as the following:
   --  ALPHA:version:field1:field2:...:fieldN
   --  What the fields mean may differ from version to version, so we're going to have to check for that.
   -- Since proper versioning has not begun yet, we will omit version checks for now.
   --local minor,major = string.format(settings[2], "(%d+).(%d+)")
   if set[3] then cfg.award = set[3] end
   if set[4] then cfg.freq = set[4] end
   if set[5] then cfg.cost = set[5] end
   if set[6] then cfg.maxlp = set[6] end
end

local function guildCfgSet()
   if not IsInGuild() or not CanEditGuildInfo() then
      ALPHA:Print("I'm sorry, Dave. I'm afraid I can't do that.")
      return
   end

   local _,major,minor = getVersion()
   local guildInfoText = GetGuildInfoText()
   local i,j = string.find(guildInfoText, "ALPHA:(%d+).(%d+):([%d:]+):END")
   local setStr = "ALPHA:"..major.."."..minor..":"..cfg.award..":"..cfg.freq..":"..cfg.cost..":"..cfg.maxlp..":END"
   if not i then
      guildInfoText = guildInfoText.."\n"..setStr
   else
      guildInfoText = string.sub(guildInfoText, 1, i-1)..setStr..string.sub(guildInfoText, j+1)
   end
   
   SetGuildInfoText(guildInfoText)
   SendAddonMessage("ALPHA", "G" .. SEP .. major .. "." .. minor .. SEP .. cfg.award .. SEP .. cgf.freq .. SEP .. cfg.cost .. SEP .. cfg.maxlp, "GUILD")
end

-- Initialization and related stuff.

function ALPHA:OnDisable()
   --stuff
end

function ALPHA:OnEnable()
   self:RegisterChatCommand("alpha", "ChatCommand")

   self:RegisterEvent("CHAT_MSG_ADDON", "Comm")
   self:RegisterEvent("LOOT_CLOSED", "LootClosed")
   self:RegisterEvent("LOOT_OPENED", "LootCheck")
   self:RegisterEvent("LOOT_SLOT_CHANGED", "LootSlotChanged")
   self:RegisterEvent("LOOT_SLOT_CLEARED", "LootSlotCleared")
--   self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED", "GroupCheck") -- Apart from the obvious, also fires when joining a group and when the threshold is changed. Basically, we can use this to detect any relevant changes apart from leaving a group.
   -- ^ Keeping track of the loot method may, interestingly enough, be kind of redundant.
   if not RegisterAddonMessagePrefix("ALPHA") then self:Print("Could not register for addon communication.") end
   if IsInGuild() then
      SendAddonMessage("ALPHA", "V" .. SEP .. GetAddOnMetadata("ALPHA", "Version"), "GUILD")
      guildCfgGet()
      for k,v in ipairs(guildInfotext) do if string.match(v, "^ALPHA:(.+)") then guildCfgGet(string.gsub(v, ":", "\7")) end end
   end
   if IsInGroup() then SendAddonMessage("ALPHA", "Y", "RAID") end
end

function ALPHA:OnInitialize()
   -- set up the gui
   gui = CreateFrame("frame", nil, UIParent)
   gui:SetPoint("TOPLEFT", cfg.private.anchor.x, cfg.private.anchor.y)
   gui:SetMovable(true)
   gui:SetResizable(false)
   gui:SetSize(500, 500)
   gui:SetFrameStrata("MEDIUM")
   gui:Hide()

   local backdrop = { -- using the same for all child frames anyway
      bgFile = [[Interface\Tooltips\UI-Tooltip-Background]],
      edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
      tile = true,
      tileSize = 32,
      edgeSize = 16,
      insets = {
	 left = 5,
	 right = 5,
	 top = 5,
	 bottom = 5,
      }
   }

   -- info frame
   gui.info = CreateFrame("frame", nil, gui)
   gui.info:SetPoint("TOPLEFT", gui, "TOPLEFT", 0, 0)
   gui.info:SetPoint("TOPRIGHT", gui, "TOPRIGHT", 0, 0)
   gui.info:SetSize(500, 50)
   gui.info:SetBackdrop(backdrop)
   gui.info:SetBackdropColor(0, 0, 0, .8)

   gui.info.current_title = gui.info:CreateFontString(nil, "MEDIUM")
   gui.info.current_title:SetPoint("TOPLEFT", 5, -5)
   gui.info.current_title:SetSize(95, 20)
   gui.info.current_title:SetFontObject("GameTooltipTextSmall")
   gui.info.current_title:SetText("Currently rolling:")
   gui.info.current_item = gui.info:CreateFontString(nil, "MEDIUM")
   gui.info.current_item:SetPoint("TOPLEFT", 100, -5)
   gui.info.current_item:SetSize(95, 20)
   gui.info.current_item:SetFontObject("GameTooltipTextSmall")
   gui.info.current_item:SetText("N/A")
   gui.info.current_item:SetJustifyH("RIGHT")

   gui.info.chance_title = gui.info:CreateFontString(nil, "MEDIUM")
   gui.info.chance_title:SetPoint("TOPLEFT", 5, -25)
   gui.info.chance_title:SetSize(95, 20)
   gui.info.chance_title:SetFontObject("GameTooltipTextSmall")
   gui.info.chance_title:SetText("Chance to win")
   gui.info.chance_info = gui.info:CreateFontString(nil, "MEDIUM")
   gui.info.chance_info:SetPoint("TOPLEFT", 100, -25)
   gui.info.chance_info:SetSize(95, 20) -- quite large for something that's just a percentage, eh? gonna have to do lots of tweaks to this part of the gui later, just doing it do actually have the frames set up for now.
   gui.info.chance_info:SetFontObject("GameTooltipTextSmall")
   gui.info.chance_info:SetText("N/A")
   gui.info.chance_info:SetJustifyH("RIGHT")

   gui.info.lpamount = CreateFrame("EditBox", nil, gui.info, "InputBoxTemplate")
   gui.info.lpamount:SetPoint("TOPLEFT", 400, -5)
   gui.info.lpamount:SetSize(45, 20)
   gui.info.lpamount:SetText(cfg.guild.award)
   gui.info.lpamount:SetAutoFocus(false)
   gui.info.lpamount:SetScript("OnEscapePressed", function() gui.info.lpamount:ClearFocus() end)
   gui.info.lpamount:SetScript("OnEnterPressed", function() gui.info.lpamount:ClearFocus() end)
   gui.info.lpaward = CreateFrame("Button", nil, gui.info, "UIPanelButtonTemplate")
   gui.info.lpaward:SetPoint("TOPLEFT", 450, -5)
   gui.info.lpaward:SetSize(45, 20)
   gui.info.lpaward:SetText("Award")
   gui.info.lpaward:SetScript("OnClick", function(self, button, down)
				 if not CanEditOfficerNote() then
				    ALPHA:Print("I'm sorry, I can't let you do that.")
				    self:Disable()
				 end
				 for i = 1, GetNumGroupMembers() do
				    local name = GetRaidRosterInfo(i)
				    if UnitIsInMyGuild(name) then
				       if not modLuckPoints(name, gui.info.lpamount:GetText()) then ALPHA:Print("Could not modify luck points for", name) end
				    end
				 end
   end)

   -- roll frame
   gui.roll = CreateFrame("frame", nil, gui)
   gui.roll:SetSize(250, 20)
   gui.roll:SetPoint("TOPLEFT", gui.info, "BOTTOMLEFT", 0, 0)
--   gui.roll:SetBackdrop(backdrop)
--   gui.roll:SetBackdropColor(0, 0, 0, .8)
   gui.roll.rec = {}
   gui.roll.cnt = {}
   gui.roll:SetScript("OnSizeChanged", function(self, width, height)
			 gui:SetHeight(gui.info:GetHeight() + (height > gui.loot:GetHeight() and height or gui.loot:GetHeight()))
   end)

   gui.roll.index = CreateFrame("frame", nil, gui.roll)
   gui.roll.index:SetBackdrop(backdrop)
   gui.roll.index:SetBackdropColor(0, 0, 0, .8)
   gui.roll.index:SetPoint("TOPLEFT", 0, 0)
   gui.roll.index:SetSize(250, 20)
   gui.roll.index.name = CreateFrame("frame", nil, gui.roll.index)
   gui.roll.index.name:SetPoint("TOPLEFT", 0, 0)
   gui.roll.index.name:SetSize(130, 20)
   gui.roll.index.name.text = gui.roll.index.name:CreateFontString(nil, "MEDIUM")
   gui.roll.index.name.text:SetFontObject("GameTooltipTextSmall")
   gui.roll.index.name.text:SetPoint("LEFT", 10, 0)
   gui.roll.index.name.text:SetSize(110, 20)
   gui.roll.index.name.text:SetText("Roller")
   
   gui.roll.index.roll = CreateFrame("frame", nil, gui.roll.index)
   gui.roll.index.roll:SetPoint("TOPLEFT", 130, 0)
   gui.roll.index.roll:SetSize(40, 20)
   gui.roll.index.rollTex = gui.roll.index.roll:CreateTexture(nil, "BACKGROUND")
   gui.roll.index.rollTex:SetTexture([[Interface\BUTTONS\UI-GroupLoot-Dice-Up]])
   gui.roll.index.rollTex:SetPoint("TOPRIGHT", -5, -2)
   gui.roll.index.rollTex:SetSize(16, 16)

   gui.roll.index.lp = CreateFrame("frame", nil, gui.roll.index)
   gui.roll.index.lp:SetPoint("TOPLEFT", 170, 0)
   gui.roll.index.lp:SetSize(40, 20)
   gui.roll.index.lpTex = gui.roll.index.lp:CreateTexture(nil, "BACKGROUND")
   gui.roll.index.lpTex:SetTexture([[Interface\BUTTONS\UI-GroupLoot-Coin-Up]]) -- Alt Interface\MINIMAP\TRACKING\StableMaster
   gui.roll.index.lpTex:SetPoint("TOPRIGHT", -5, -2)
   gui.roll.index.lpTex:SetSize(16, 16)

   gui.roll.index.sum = CreateFrame("frame", nil, gui.roll.index)
   gui.roll.index.sum:SetPoint("TOPLEFT", 210, 0)
   gui.roll.index.sum:SetSize(40, 20)
   gui.roll.index.sumTex = gui.roll.index.sum:CreateTexture(nil, "BACKGROUND")
   gui.roll.index.sumTex:SetTexture([[Interface\MINIMAP\TRACKING\StableMaster]]) -- Replace with actually getting a texture of the sum symbol. (greek big sigma)
   gui.roll.index.sumTex:SetPoint("TOPRIGHT", -5, -2)
   gui.roll.index.sumTex:SetSize(16, 16)
   
   function gui.roll:New(tbl)
      local rgui
      if #self.rec > 0 then rgui = table.remove(self.rec)
      else
	 rgui = CreateFrame("frame", nil, self)
	 rgui:SetSize(250, 20)
	 rgui:SetBackdrop(backdrop)
	 rgui:SetBackdropColor(0, 0, 0, .8) -- Maybe, I dunno. Just hacking off the loot gui atm. >.>

	 rgui.name = CreateFrame("frame", nil, rgui)
	 rgui.name:SetPoint("TOPLEFT", 0, 0)
	 rgui.name:SetSize(130, 20)
	 rgui.name.text = rgui.name:CreateFontString(nil, "MEDIUM")
	 rgui.name.text:SetFontObject("GameToolTipTextSmall")
	 rgui.name.text:SetPoint("LEFT", 10, 0)
	 rgui.name.text:SetSize(110, 20)

	 rgui.roll = CreateFrame("frame", nil, rgui)
	 rgui.roll:SetPoint("TOPLEFT", 130, 0)
	 rgui.roll:SetSize(40, 20)
	 rgui.roll.text = rgui.roll:CreateFontString(nil, "MEDIUM")
	 rgui.roll.text:SetFontObject("GameToolTipTextSmall")
	 rgui.roll.text:SetPoint("RIGHT", -5, 0)
	 rgui.roll.text:SetSize(30, 20)
	 rgui.roll.text:SetJustifyH("RIGHT")

	 rgui.lp = CreateFrame("frame", nil, rgui)
	 rgui.lp:SetPoint("TOPLEFT", 170, 0)
	 rgui.lp:SetSize(40, 20)
	 rgui.lp.text = rgui.lp:CreateFontString(nil, "MEDIUM")
	 rgui.lp.text:SetFontObject("GameTooltipTextSmall")
	 rgui.lp.text:SetPoint("RIGHT", -5, 0)
	 rgui.lp.text:SetSize(30, 20)
	 rgui.lp.text:SetJustifyH("RIGHT")

	 rgui.sum = CreateFrame("frame", nil, rgui)
	 rgui.sum:SetPoint("TOPLEFT", 210, 0)
	 rgui.sum:SetSize(40, 20)
	 rgui.sum.text = rgui.sum:CreateFontString(nil, "MEDIUM")
	 rgui.sum.text:SetFontObject("GameTooltipTextSmall")
	 rgui.sum.text:SetPoint("RIGHT", -5, 0)
	 rgui.sum.text:SetSize(30, 20)
	 rgui.sum.text:SetJustifyH("RIGHT")
      end

      rgui.tbl = tbl
      rgui.name.text:SetText(tbl.roller)
      local class,cindex = UnitClass(tbl.roller)
      if cindex then
	 rgui.name.text:SetTextColor(RAID_CLASS_COLORS[cindex].r, RAID_CLASS_COLORS[cindex].g, RAID_CLASS_COLORS[cindex].b, 1)
      else
	 rgui.name.text:SetTextColor(.7, .7, .7, 1)
      end
      rgui.roll.text:SetText(tbl.roll)
      local _,r,g,b = Prism:Gradient("hsv", 1, 0, 0, 1, 0, 0, tbl.roll / 100)
      rgui.roll.text:SetTextColor(r, g, b, 1)
      rgui.lp.text:SetText(tbl.lp)
      rgui.sum.text:SetText(tbl.roll + tbl.lp) -- Should hopefully work, given how lua handles these things. Might have to rewrite, though, because it might end up tossing an integer argument where string should be expected.

      rgui.Update = self.UpdateChild
      table.insert(self.cnt, rgui)
      self:Update()
      rgui:Show()
      
      return rgui
   end

   function gui.roll:UpdateChild(highSum, highLP)
      local _,r,g,b
      if useLP then _,r,g,b = Prism:Gradient("hsv", 1, 0, 0, 1, 0, 0, self.tbl.lp / highLP)
      else r,g,b = 100,100,100 end -- Not sure if this is a nice gray or not. Or is it grey? Which is the non-US again? Damnit.
      self.lp.text:SetTextColor(r, g, b, 1)
      if useLP then _,r,g,b = Prism:Gradient("hsv", 1, 0, 0, 1, 0, 0, (self.tbl.roll + self.tbl.lp) / highSum)
      else _r,g,b = 100,100,100 end
      self.sum.text:SetTextColor(r, g, b, 1)
      -- Now I would also set the backdrop to this color, only it seems rather stupid to set the backdrop to the same color as one of the fields. Just keep it black for proper visibility, mkay?
   end

   function gui.roll:Update()
      self:SetHeight((#self.cnt + 1) * 20)

      if useLP then
	 table.sort(self.cnt, function(a, b) return (a.tbl.roll + a.tbl.lp) > (b.tbl.roll + b.tbl.lp) end) -- might be better to ensure it's sorted exactly the same as the 'rolls' table. And for equal sums, this sorting could indeed end up different, even though the sorting algorithm remains the same. Because lua works like this.
      else
	 table.sort(self.cnt, function(a, b) return a.tbl.roll > b.tbl.roll end)
      end

      -- Fetching the highs here and passing them as arguments to the update function for individual roll frames, rather than iterating over all roll data quite needlessly. And yes, all this just because COSMETICS!
      local highSum = 0
      local highLP = 0

      for k,v in ipairs(rolls) do
	 if v.lp > highLP then highLP = v.lp end
	 if (v.roll + v.lp) > highSum then highSum = (v.roll + v.lp) end -- why am I even using parentheses?
      end
      
      for i = 1, #self.cnt do
	 self.cnt[i]:SetPoint("TOPLEFT", 0, i * -20)
	 self.cnt[i]:Update(highSum, highLP)
      end
   end
	 
   
   -- loot frame
   gui.loot = CreateFrame("frame", nil, gui)
   gui.loot:SetWidth(250)
   gui.loot:SetPoint("TOPRIGHT", gui.info, "BOTTOMRIGHT", 0, 0)
--   gui.loot:SetBackdrop(backdrop)
--   gui.loot:SetBackdropColor(0, 0, 0, .8)
   gui.loot.rec = {}
   gui.loot.cnt = {}
   gui.loot:SetScript("OnSizeChanged", function(self, width, height)
			 gui:SetHeight(gui.info:GetHeight() + (height > gui.roll:GetHeight() and height or gui.roll:GetHeight()))
   end)

   function gui.loot:New(tbl)
      local lgui
      if #self.rec > 0 then lgui = table.remove(self.rec)
      else
	 lgui = CreateFrame("frame", nil, self)
	 lgui:SetSize(250, 50)
	 lgui:SetBackdrop(backdrop)
	 lgui:SetBackdropColor(0, 0, 0, .8)

	 lgui.icon = CreateFrame("frame", nil, lgui)
	 lgui.icon:SetPoint("TOPLEFT", 5, -5)
	 lgui.icon:SetSize(40, 40)
	 lgui.iconTex = lgui.icon:CreateTexture(nil, "BACKGROUND")
	 lgui.iconTex:SetAllPoints()
	 lgui.icon.count = lgui.icon:CreateFontString(nil, "OVERLAY")
	 lgui.icon.count:SetPoint("BOTTOMRIGHT", -2, 2)
	 lgui.icon.count:SetFont(GameFontNormal:GetFont(), 16, "OUTLINE")

	 lgui.icon.warforged = lgui.icon:CreateFontString(nil, "OVERLAY")
	 lgui.icon.warforged:SetPoint("TOPLEFT", 2, -2)
	 lgui.icon.warforged:SetFont(GameFontNormal:GetFont(), 14, "OUTLINE")
	 lgui.icon.warforged:SetText("W")
	 lgui.icon.warforged:SetTextColor(1, 0, 0, 1)

	 lgui.icon.tertiary = lgui.icon:CreateFontString(nil, "OVERLAY")
	 lgui.icon.tertiary:SetPoint("BOTTOMLEFT", 2, 2)
	 lgui.icon.tertiary:SetFont(GameFontNormal:GetFont(), 14, "OUTLINE")
	 lgui.icon.tertiary:SetTextColor(1, 1, 0, 1)

	 lgui.icon.socket = lgui.icon:CreateFontString(nil, "OVERLAY")
	 lgui.icon.socket:SetPoint("TOPRIGHT", -2, -2)
	 lgui.icon.socket:SetFont(GameFontNormal:GetFont(), 14, "OUTLINE")
	 lgui.icon.socket:SetText("S")
	 lgui.icon.socket:SetTextColor(0, 1, 0, 1)

--	 lgui.icon.warforged:SetTexture([[Interface\BUTTONS\UI-GroupLoot-DE-Up]])
--	 lgui.icon.warforged:SetPoint("TOPLEFT", 2, -2)
--	 lgui.icon.warforged:SetSize(16, 16)

--	 lgui.icon.socket = loot.icon:CreateTexture(nil, "OVERLAY")
--	 lgui.icon.socket:SetTexture([[Interface\ITEMSOCKETINGFRAME\UI-EmptySocket-Prismatic]])
--	 lgui.icon.socket:SetPoint("TOPRIGHT", -2, -2)
--	 lgui.icon.socket:SetSize(16, 16)

--	 lgui.icon.tertiary = loot.icon:CreateTexture(nil, "OVERLAY")
--	 lgui.icon.tertiary:SetTexture([[Interface\BUTTONS\UI-GroupLoot-DE-Up]])
--	 lgui.icon.tertiary:SetPoint("BOTTOMLEFT", 2, 2)
--	 lgui.icon.tertiary:SetSize(16, 16)

	 ---[[Interface\BUTTONS\UI-GroupLoot-Dice-Up]]
-- GossipFrame\TaxiGossipIcon

--MINIMAP\TRACKING\FlightMaster -- Speed?
--MINIMAP\TRACKING\Repair -- Indest?
--MINIMAP\TRACKING\Poisons -- Leech?
--TAXIFRAME\UI-Taxi-Icon-Green -- Avoid?

--ICONS\Ability_Defend -- Indest?
--ICONS\Ability_DeathKnight-DeathSiphon
--ICONS\Ability_Rogue_Sprint

	 lgui.name = CreateFrame("frame", nil, lgui)
	 lgui.name:SetPoint("TOPLEFT", 45, -5)
	 lgui.name:SetSize(125, 40)
	 lgui.name:SetBackdrop(backdrop)
	 lgui.name:SetBackdropColor(0, 0, 0, .8)
	 lgui.name.text = lgui.name:CreateFontString(nil, "MEDIUM")
	 lgui.name.text:SetFontObject("GameTooltipTextSmall")
	 lgui.name.text:SetPoint("LEFT", 10, 0)
	 lgui.name.text:SetSize(105, 30)

	 lgui.mlbtn = CreateFrame("BUTTON", nil, lgui, "UIPanelButtonTemplate")
	 lgui.mlbtn:SetPoint("TOPLEFT", 170, -5)
	 lgui.mlbtn:SetSize(75, 20)
	 lgui.mlbtn:SetScript("OnClick", function(self, button, down)
				 local parent = self:GetParent() -- Doubt self:GetParent().tbl.iid would go well.
				 if not current_roll then
				    ALPHA:StartRoll(parent.tbl.iid, cfg.private.rolltime, UnitName("player"), true)
				 elseif current_roll == parent.tbl.iid then
				    ALPHA:AwardLoot(parent.tbl.iid)
				 else
				    ALPHA:Print("There is already a roll in progress. Please wait for it to finish.")
				 end
	 end)

	 lgui.queue = CreateFrame("BUTTON", nil, lgui, "UIPanelButtonTemplate")
	 lgui.queue:SetPoint("TOPLEFT", 170, -25)
	 lgui.queue:SetSize(75, 20)
	 lgui.queue:SetScript("OnClick", function(self, button, down)
				 local parent = self:GetParent()

				 if not current_roll or current_roll ~= parent.tbl then
				    parent.tbl.queued = true
				 else
				    RandomRoll(1, 100)
				    current_roll.rolled = true
				 end
				 parent:Update()
	 end)
				    
	 lgui.Update = self.UpdateChild
      end

      lgui.tbl = tbl
      lgui.iconTex:SetTexture(GetItemIcon(tbl.id))
      lgui.name.text:SetText("|c" .. tbl.color .. tbl.name .. "|r") -- Kind of ugly way of doing it. But kind of the smoothest.
      lgui.name.text:SetTextColor()
      if tbl.warforged then lgui.icon.warforged:Show() else lgui.icon.warforged:Hide() end
      if tbl.socket then lgui.icon.socket:Show() else lgui.icon.socket:Hide() end
      if tbl.tertiary then
	 if tbl.tertiary == 40 then lgui.icon.tertiary:SetText("A") -- Avoidance
	 elseif tbl.tertiary == 41 then lgui.icon.tertiary:SetText("L") -- Leech
	 elseif tbl.tertiary == 42 then lgui.icon.tertiary:SetText("S") -- Speed
	 elseif tbl.tertiary == 43 then lgui.icon.tertiary:SetText("I") end -- Indestruct
	 lgui.icon.tertiary:Show()
      else lgui.icon.tertiary:Hide() end

      table.insert(self.cnt, lgui)
      self:Update()
      lgui:Show()
      return lgui
   end

   function gui.loot:UpdateChild()
      if lootOpen then
	 if #self.tbl.slots > 0 then
	    if not current_roll then
	       self.mlbtn:SetText("Start Rolls")
	       self.mlbtn:Enable()
	    elseif current_roll ~= self.tbl then
	       self.mlbtn:SetText("Start Rolls")
	       self.mlbtn:Disable()
	    elseif not current_roll.done then -- rolling
	       self.mlbtn:SetText("In Progress")
	       self.mlbtn:Disable()
	    else  -- finished, waiting to distribute
	       self.mlbtn:SetText("Award")
	       self.mlbtn:Enable()
	    end
	 else
	    local rnd = math.random(1, #blargh)
	    self.mlbtn:SetText(blargh[rnd])
	    self.mlbtn:Disable()
	 end

	 if current_roll and current_roll == self.tbl then
	    if current_roll.rolled then
	       self.queue:SetText("Rolled")
	       self.queue:Disable()
	    else
	       self.queue:SetText("Roll")
	       self.queue:Enable()
	    end
	 elseif self.tbl.queued then
	    self.queue:SetText("Queued")
	    self.queue:Disable()
	 else
	    self.queue:SetText("Queue Roll")
	    self.queue:Enable()
	 end

	 local mycount,count = 0,0

	 for k,v in pairs(self.tbl.masters) do
	    if k == UnitName("player") then mycount = v
	    else count = count + v end
	 end

	 if mycount + count > 1 then
	    self.icon.count:SetText(((mycount > 0) and ("|cff00ff00" .. mycount .. "|r + ") or "") .. count)
	    self.icon.count:Show()
	 else
	    self.icon.count:Hide()
	 end
	 
      else
	 self.mlbtn:Disable()
	 self.queue:Disable()
      end
   end


   function gui.loot:Update()
      self:SetHeight(#self.cnt * 50)

      table.sort(self.cnt, function(a, b)
		    return a.tbl.iid > b.tbl.iid end) -- Should work rather well.
--		    if a.tbl.id == b.tbl.id then
--		       local aScore = 0
--		       local bScore = 0
--
--		       -- Let's just use Blizz numbers for the sorting score.
--		       if a.tbl.warforged then aScore = aScore + 561 end
--		       if b.tbl.warforged then bScore = bScore + 561 end
--		       if a.tbl.socket then aScore = aScore + 565 end
--		       if b.tbl.socket then bScore = bScore + 565 end
--		       if a.tbl.tertiary then aScore = aScore + a.tbl.tertiary end
--		       if b.tbl.tertiary then bScore = bScore + a.tbl.tertiary end
--
--		       return aScore > bScore
--		    else return a.tbl.id < b.tbl.id end
--      end)
-- Lots of comment lines because for some reason it screws with my emacs if I use the delimiter.

      for i = 1, #self.cnt do
	 self.cnt[i]:SetPoint("TOPLEFT", 0, (i - 1) * -50)
	 self.cnt[i]:Update()
      end
   end
end

-- Other crap!
function ALPHA:ChatCommand(...)
   if gui:IsShown() then gui:Hide() else gui:Show() end
end

function ALPHA:AwardLoot(iid)
   local winner
   for k,v in ipairs(rolls) do
      if not v.disabled then
	 winner = v
	 break
      end
   end
   if not winner then -- If this is the case, we probably want to have a disenchanter set and give the stuff to them instead.
      self:Print("No winner for", loot[iid].name)
      return
   end
   for k,v in ipairs(loot[iid].slots) do
      for i = 1, MAX_RAID_MEMBERS do
	 if winner.roller == GetMasterLootCandidate(v, i) then
	    GiveMasterLoot(v, i)
	    if useLP and UnitIsInMyGuild(winner.roller) then
	       if not modLuckPoints(winner.roller, cfg.guild.cost * -1) then self:Print("Failed to modify luck points for", winner.roller) end
	    end
	 end
      end
   end
   self:Print("Failed to award item, disabling eligibility for", winner.roller)
   winner.disabled = true
   -- Now we could call the function again to just go to the next in the list, but that could be bad. Not for technical reasons, really, but because something failed and people might want to have the chance to look into it rather than having the addon just chucking loot at whoever.
end

function ALPHA:AddLoot(itemLink, amount, master, lootSlot)
   --   local found = false
   local rec = false
   local iid,name,id,color,warforged,socketed,tertiary = parseLootLink(itemLink)

   if loot[iid] then
      if loot[iid].lost > 0 and lootSlot then -- Since we're only doing this for items we're the master looter for, we can ignore the fact that "amount" can be > 1, since it can't in that case.
	 loot[iid].lost = loot[iid].lost - 1
	 rec = true
      else
	 loot[iid].count = loot[iid].count + amount
	 loot[iid].gui:Update()
      end
      if loot[iid].masters[master] then loot[iid].masters[master] = loot[iid].masters[master] + amount
      else loot[iid].masters[master] = amount end
   else
      loot[iid] = {
	 id = id,
	 iid = iid, -- Yes, we have it in the key already. But sometimes we'll access it without knowing the key, and we'll want to know the key .. so uh, yeah. This seems like a better approach rather than to iterate over the table and compare in order to fetch. Or calling parseLootLink() for that matter. I mean sure, we could probably try to reverse-engineer from other data here, but that seems unsafe.
	 name = name,
	 link = itemLink,
	 color = color,
	 count = amount,
	 lost = 0,
	 warforged = warforged,
	 socket = socketed,
	 tertiary = tertiary,
	 queued = false,
	 rolled = false,
	 slots = {},
	 masters = {}
      }

      loot[iid].masters[master] = amount

      local lgui = gui.loot:New(loot[iid])
      loot[iid].gui = lgui
   end

   if lootSlot then
      loot.slots[lootSlot] = loot[iid]
      table.insert(loot[iid].slots, lootSlot)
      if not rec then SendAddonMessage("ALPHA", "L" .. SEP .. itemLink .. SEP .. "1", "RAID") end
   end
end

function ALPHA:RemoveLoot(iid, amount, master, lootSlot)
   if loot[iid] then
      if not loot[iid].masters[master] then return end
      if loot[iid].masters[master] < amount then amount = loot[iid].masters[master] end
      if lootSlot then
	 for k,v in ipairs(loot[iid].slots) do
	    if v == lootSlot then table.remove(loot[iid].slots, k) end
	 end
	 loot.slots[lootSlot] = nil
      end
      if loot[iid].count > amount then
	 self:Print("Decreasing counter for",loot[iid].name)
	 loot[iid].count = loot[iid].count - amount
	 loot[iid].gui:Update()
	 if loot[iid].masters[master] == amount then loot[iid].masters[master] = nil
	 else loot[iid].masters[master] = loot[iid].masters[master] - amount end
      else
	 self:Print("Removing",loot[iid].name)
	 if current_roll == loot[iid] then
	    gui.info.current_item:SetText("N/A")
	    current_roll = nil
	    loot[iid].masters[master] = nil
	    for i = #gui.roll.cnt, 1, -1 do -- Going backwards since we'll be removing everything.
	       gui.roll.cnt[i]:Hide()
	       gui.roll.cnt[i].tbl = nil
	       table.insert(gui.roll.rec, table.remove(gui.roll.cnt, i))
	    end
	    for i = #rolls, 1, -1 do -- Shouldn't be as important to go backwards here, since we're not calling table.remove(), but doing it anyway! JUST BECAUSE! Here, have a cookie.
	       for k,v in pairs(rolls[i]) do rolls[i][k] = nil end -- I mean, I guess we could just call wipe(), but I kinda like doing it manually like this, ok? OK? Ok. Ook ook.
	       rolls[i] = nil
	    end
	    
	 end
	 for k,v in ipairs(gui.loot.cnt) do
	    if v == loot[iid].gui then
	       v:Hide()
	       table.insert(gui.loot.rec, table.remove(gui.loot.cnt, k))
	       break
	    end
	    loot[iid].gui.tbl = nil
	 end
	 
	 for k,v in pairs(loot[iid]) do loot[iid][k] = nil end
	 loot[iid] = nil

	 gui.loot:Update()
      end
   end
end
	 
-- More stuffings

function ALPHA:StartRoll(iid, time, masterLooter, button)
   if button then -- We're the person handling this item.
      self:ScheduleTimer("RollTimer", time) -- For now. Documentation is wonky, have to actually check the AceTimer code to figure out how to use it best.
      SendAddonMessage("ALPHA", "S" .. SEP .. iid .. SEP .. time, "RAID")
      self:RegisterEvent("CHAT_MSG_SYSTEM", "CheckRoll") -- May have to move this in case we need to use it to detect players leaving a group as well.
      
      local function buildCandidateList(slot, index)
	 local candidate = GetMasterLootCandidate(slot, index)
	 if not candidate then return end
	 if candidate_list[candidate] then candidate_list[candidate].slots[slot] = true
	 else
	    local lp
	    if IsInGuild() and UnitIsInMyGuild(candidate) and CanViewOfficerNote() then -- Need to move the CanViewOfficerNote() to somewhere else. Give the user a chance to .. something.
	       lp = getLuckPoints(candidate)
	    else
	       lp = 0
	    end
	    candidate_list[candidate] = {
	       lp = lp,
	       slots = {}
	    }
	    candidate_list[candidate].slots[slot] = true
	 end
	 buildCandidateList(slot, index + 1)
      end
      
      for k,v in pairs(loot[iid].slots) do buildCandidateList(v, 1) end
	 
   end

   if IsInGuild() and UnitIsInMyGuild(masterLooter) then
      myLP = getLuckPoints(UnitName("player"))
   else
      myLP = 0
   end
   
   current_roll = loot[iid]
   if current_roll.queued then
      RandomRoll(1, 100)
      current_roll.rolled = true
   end

   gui.info.current_item:SetText("|c" .. current_roll.color .. current_roll.name .. "|r")
   gui.loot:Update() -- Because buttons.

   if DBM then
      DBM.Bars:CreateBar(tonumber(time), "Loot Roll: " .. current_roll.name, GetItemIcon(current_roll.id))
   end
end

function ALPHA:EndRoll(isMaster)
   if isMaster then
      SendAddonMessage("ALPHA", "T")
      self:UnregisterEvent("CHAT_MSG_SYSTEM")
   end

   current_roll.done = true

   -- if the item is removed midroll, the below will toss an error, so gotta accomodate for that. probably by hacking an EndRoll() call into RemoveLoot() (or whatever I named that)
   current_roll.gui:Update()
end

function ALPHA:RollTimer()
   self:EndRoll(true)
end

function ALPHA:AddRoll(roller, roll, lp)
   table.insert(rolls, {roller = roller, roll = tonumber(roll), lp = tonumber(lp)})
   local tbl = rolls[#rolls] -- So as to not lose it post-sorting.
   table.sort(rolls, function(a, b)
		 if useLP then return (a.roll + a.lp) > (b.roll + b.lp)
		 else return a.roll > b.roll end
   end)

   local rgui = gui.roll:New(rolls[#rolls])
   tbl.gui = rgui -- not sure we need it. we might, though. so keep this for now.

   local chance = useLP and ((100 + myLP) - (rolls[1].roll + rolls[1].lp)) or (100 - rolls[1].roll)
   if chance < 0 then chance = 0 elseif chance > 100 then chance = 100 end
   local _,r,g,b = Prism:Gradient("hsv", 1, 0, 0, 1, 0, 0, chance/100)
   gui.info.chance_info:SetText(chance .. "%")
   gui.info.chance_info:SetTextColor(r, g, b, 1)
end

-- Event handlers
function ALPHA:CheckRoll(_, msg)
   if string.match(msg, RANDOM_ROLL_RESULT) then
      local roller, roll, rMin, rMax = string.match(msg, RANDOM_ROLL_RESULT)
      if not candidate_list[roller] then
	 SendChatMessage("You are not a valid loot candidate for " .. current_roll.link, "WHISPER", nil, roller)
	 self:Print(roller, "is not a valid loot candidate for this item, apparently.")
      else
	 local found = false
	 for k,v in ipairs(rolls) do
	    if roller == v.name then
	       found = true
	       break
	    end
	 end
	 if not found then -- Allowing people to reroll would be silly.
	    self:AddRoll(roller, roll, candidate_list[roller].lp)
	    SendAddonMessage("ALPHA", "R" .. SEP .. roller .. SEP .. roll .. SEP .. candidate_list[roller].lp)
	 end
      end
   end
end

function ALPHA:Comm(_, prefix, message, channel, sender)
   local name,realm = UnitFullName("player")
   if prefix ~= "ALPHA" or sender == name.."-"..realm then return end

   local args = {strsplit(SEP, message)}

   -- For the below comments, \7 is substituted with ":" for readability.
   
   if args[1] == "D" then
      -- Remove ([d]elete) a tracked item
      -- D:internal id:amount
      self:RemoveLoot(args[2], args[3], sender)

   elseif args[1] == "G" then
      -- [G]uild settings changed
      -- G:version:fields (see guildCfgGet() for details)
      guildCfgGet(message)

   elseif args[1] == "L" then
      -- [L]oot
      -- L:itemlink:amount
      self:AddLoot(args[2], args[3], sender)

--   elseif args[1] == "P" then -- Use (p)oints
--      if args == "0" then useLP = false else useLP = true end

   elseif args[1] == "R" then
      -- [R]oll
      -- R:roller:roll:luck points
      self:AddRoll(args[2], args[3], args[4])

   elseif args[1] == "S" then
      -- [S]tart a roll
      -- S:internal id:time in seconds
      self:StartRoll(args[2], args[3], sender, false)

   elseif args[1] == "T" then -- S(t)op a roll
      -- S[t]op a roll
      -- T
      self:EndRoll()

   elseif args[1] == "V" then
      -- [V]ersion check
      -- V:version string
      local mymajor,myminor
      if update_found then mymajor,myminor = string.match(update_found, "(%d+).(%d+)")
      else _,mymajor,myminor = getVersion() end
      if not mymajor then mymajor = 0 end
      if not myminor then myminor = 0 end
      local major,minor = string.match(args[2], "(%d+).(%d+)")
      -- in case we get something weird
      if not major then major = 0 end
      if not minor then minor = 0 end
      major,minor,mymajor,myminor = tonumber(major),tonumber(minor),tonumber(mymajor),tonumber(myminor)
      if major > mymajor or (major == mymajor and minor > myminor) then
	 self:Print("Version " .. args[2] .. " is available.")
	 update_found = args[2]
      elseif major < mymajor or (major == mymajor and minor < myminor) then SendAddonMessage("ALPHA", "V" .. SEP .. GetAddOnMetadata("ALPHA", "Version"), channel) end

   elseif args[1] == "Y" then
      for k,v in pairs(loot) do
	 if #v.slots > 0 then SendAddonMessage("ALPHA", "L" .. SEP .. v.link .. SEP .. #v.slots, "WHISPER", sender) end
      end
   end

end

function ALPHA:LootClosed(event)
   lootOpen = false
   self:Print("Loot window closed.")
   for k,v in pairs(loot) do
      self:Print("Processing",k,v.name)
      if k ~= "slots" then
	 self:Print(#v.slots, "slots tracked")
	 v.lost = #v.slots
	 self:Print("Lost",v.lost,v.link)
	 for s,ml in pairs(v.slots) do v[s] = nil end
      end
   end
   loot.slots = {}
end

function ALPHA:LootCheck(event)
   lootOpen = true
--   if not IsInGroup() then return end
--[[   local function buildCandidateList(slot, index)
      local candidate = GetMasterLootCandidate(slot, index)
      if not candidate then return end
      if IsInGuild() and UnitIsInMyGuild(candidate) and CanViewOfficerNote() then
	 local lp = getLuckPoints(candidate)
	 loot[slot].candidates[candidate] = lp
      else
	 loot[slot].candidates[candidate] = 0
      end
      buildCandidateList(slot, index + 1)
   end
]]--

   for i = 1, GetNumLootItems() do
      if LootSlotHasItem(i) and GetMasterLootCandidate(i, 1) then -- Should return if we are the master looter. Sadly, it still returns for crap items below the threshold set at kill. We'll have to check against the current threshold below and hope it hasn't changed post-kill. Having changed master looter should still be safe, though.
	 local _,lootName,_,lootQuality = GetLootSlotInfo(i)
	 if lootQuality >= GetLootThreshold() then -- Yup, let's just hope this hasn't changed. Optionally hard-set desired threshold in the addon.
	    local lootLink = GetLootSlotLink(i)

	    if not lootLink then self:Print("Missing itemlink for lootslot", i, ". Skipping.")
	    else self:AddLoot(lootLink, 1, UnitName("player"), i) end
	 end
      end
   end

   for k,v in pairs(loot) do
      if k ~= "slots" and v.lost > 0 then
	 self:RemoveLoot(k, v.lost, UnitName("player"))
	 SendAddonMessage("ALPHA", "D" .. sep .. k .. v.lost)
	 v.lost = 0
      end
   end
end

function ALPHA:LootSlotChanged(event, slot) -- might want to do something here, just not sure what ..
end

function ALPHA:LootSlotCleared(event, slot)
   if loot.slots[slot] and loot[loot.slots[slot]] and loot[loot.slots[slots]].masters[UnitName("player")] then
      SendAddonMessage("ALPHA", "D" .. SEP .. loot[loot.slots[slot]].iid .. SEP .. slot)
      self:RemoveLoot(loot[loot.slots[slot]].iid, 1, UnitName("player"), slot)
   end
end

function ALPHA:EventTest(...)
   self:Print(...) -- ...
end
